/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.core;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.core.CoreObjectFactory.ActuatorToConfigure;
import de.unibi.airobots.bonsai.core.CoreObjectFactory.CoreObjectCreationException;
import de.unibi.airobots.bonsai.core.CoreObjectFactory.InitializationException;
import de.unibi.airobots.bonsai.core.CoreObjectFactory.SensorToConfigure;
import de.unibi.airobots.bonsai.core.configuration.ConfigurationParser;
import de.unibi.airobots.bonsai.core.configuration.ConfiguredActuator;
import de.unibi.airobots.bonsai.core.configuration.ConfiguredCoreObject;
import de.unibi.airobots.bonsai.core.configuration.ConfiguredCoreObjectFactory;
import de.unibi.airobots.bonsai.core.configuration.ConfiguredSensor;
import de.unibi.airobots.bonsai.core.configuration.XmlConfigurationParser;
import de.unibi.airobots.bonsai.core.configuration.ConfigurationParser.ParseException;
import de.unibi.airobots.bonsai.xcf.configuration.ConfiguredXcfCoreObject;

/**
 * Central access point to the Bonsai classes. This class provides various
 * methods to create {@link Sensor}s and {@link Actuator}s. The general usage
 * pattern is to first configure the factory via a user-provided configuration
 * file. The default parser is XML-based but can be changed if desired.
 * Afterwards the configured {@link Sensor}s and {@link Actuator}s can be
 * retrieved via the <code>create...</code> methods. A {@link Sensor} itself is
 * not thread-safe if not stated otherwise in the implementing class. The same
 * counts for actuators.
 * 
 * This class is implemented as a singleton.
 * 
 * @author sebschne
 * @author nkoester
 * @author jwienke
 */
public class BonsaiManager {

    /**
     * Exception indicating an error in the Bonsai core configuration.
     * 
     * @author jwienke
     */
    public class ConfigurationException extends RuntimeException {

        /**
         * Serial version UID.
         */
        private static final long serialVersionUID = -8111248937260206718L;

        /**
         * Constructor.
         */
        public ConfigurationException() {
            super();
        }

        /**
         * Constructor.
         * 
         * @param message
         *            exception description
         * @param cause
         *            cause of the exception
         */
        public ConfigurationException(String message, Throwable cause) {
            super(message, cause);
        }

        /**
         * Constructor.
         * 
         * @param message
         *            exception description
         */
        public ConfigurationException(String message) {
            super(message);
        }

        /**
         * Constructor.
         * 
         * @param cause
         *            cause of the exception
         */
        public ConfigurationException(Throwable cause) {
            super(cause);
        }

    }

    private Logger logger = Logger.getLogger(getClass());

    /**
     * The parser instance used for initial configuration.
     */
    private ConfigurationParser parser = new XmlConfigurationParser();

    /**
     * A map of configured factories indexed by their class.
     */
    private Map<Class<? extends CoreObjectFactory>, CoreObjectFactory> configuredFactoriesByClass =
            new HashMap<Class<? extends CoreObjectFactory>, CoreObjectFactory>();

    /**
     * Maps every configured sensor by key to its creating factory.
     */
    private Map<String, CoreObjectFactory> sensorKeysToFactories =
            new HashMap<String, CoreObjectFactory>();

    /**
     * Maps every configured actuator by key to its creating factory.
     */
    private Map<String, CoreObjectFactory> actuatorKeysToFactories =
            new HashMap<String, CoreObjectFactory>();

    /**
     * Singleton instance variable.
     */
    private static BonsaiManager instance = new BonsaiManager();

    /**
     * Private constructor for singleton pattern.
     */
    private BonsaiManager() {
    }

    /**
     * Only for testing!!! Don't use this!
     */
    static void testConstructNewInstance() {
        instance = new BonsaiManager();
    }

    /**
     * Returns the singleton instance of this class.
     * 
     * @return one and only instance of this class
     */
    public static BonsaiManager getInstance() {
        return instance;
    }

    /**
     * Sets the parser to use for the {@link #configure(String)} call.
     * 
     * @param parser
     *            new parser to use
     */
    public void setParser(ConfigurationParser parser) {
        this.parser = parser;
    }
    
    /**
     *
     * Returns the parser instance used for initial configuration.
     *
     * @return current parser instance
     */
    public ConfigurationParser getParser() {
        return parser;
    }

    /**
     * Returns the set of configured factory classes after configuration via
     * {@link #configure(URI)}.
     * 
     * @return set of configured factory classes
     */
    public Set<Class<? extends CoreObjectFactory>> getConfiguredFactoryClasses() {
        return configuredFactoriesByClass.keySet();
    }

    /**
     * Configures the factory from a given configuration file with the parser
     * set at a call to {@link #setParser(ConfigurationParser)}. Default parser
     * is xml based. This method can only be called once per program lifetime.
     * 
     * @param configurationFilename
     *            name of the local configuration file
     * @throws IllegalStateException
     *             already configured or error while configuring
     * @throws ConfigurationException
     *             error configuring the {@link BonsaiManager}. If this is the
     *             case, this class is in an unpredictable state and cannot be
     *             used afterwards
     */
    public void configure(String configurationFilename)
            throws IllegalStateException, ConfigurationException {
        try {
            configure(new URI("file://" + configurationFilename));
        } catch (URISyntaxException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Configures the factory from a given configuration file with the parser
     * set at a call to {@link #setParser(ConfigurationParser)}. Default parser
     * is xml based. This method can only be called once per program lifetime.
     * 
     * @param configurationFile
     *            uri of the configuration file
     * @throws IllegalStateException
     *             already configured or error while configuring
     * @throws ConfigurationException
     *             error configuring the {@link BonsaiManager}. If this is the
     *             case, this class is in an unpredictable state and cannot be
     *             used afterwards
     */
    public void configure(URI configurationFile) throws IllegalStateException,
            ConfigurationException {

        logger.info("Starting configuration from uri: " + configurationFile);

        try {

            logger.info("Parsing configuration file");
            parser.parseFile(configurationFile);

            logger.info("Creating factories");
            createFactories(parser.getConfiguredFactories());

            logger.info("Initializing sensors per factory");
            initializeSensors(parser.getConfiguredSensors());

            logger.info("Initializing actuators");
            initializeActuators(parser.getConfiguredActuators());

        } catch (ParseException e) {
            throw new ConfigurationException(
                    "Error parsing the configuration file: " + e.getMessage(), e);
        } catch (IOException e) {
            throw new ConfigurationException(
                    "Error accessing the configuration file: " + e.getMessage(), e);
        }
    }

    /**
     * Indicates whether the factory can create a sensor with the unique key
     * <code>key</code> and the data type class <code>dataType</code>.
     * 
     * @param key
     *            key of the sensor to check
     * @param dataType
     *            data type of the sensor with the given key to return
     * @return <code>true</code> if there is a factory that was configured for
     *         sensors with the specified key and return data type, else
     *         <code>false</code>
     */
    public boolean canCreateSensor(String key, Class<?> dataType) {
        return sensorKeysToFactories.containsKey(key)
                && sensorKeysToFactories.get(key)
                        .canCreateSensor(key, dataType);
    }

    /**
     * Tests if an actuator of the desired class can be created. The desired
     * class is interpreted as the minimally fulfilled interface.
     * 
     * @param key
     *            key of the actuator
     * @param actuatorClass
     *            minimal interface an actuator under the given key must
     *            implement
     * @return <code>true</code> if such an actuator can be created, else
     *         <code>false</code>
     */
    public boolean canCreateActuator(String key,
            Class<? extends Actuator> actuatorClass) {

        logger.debug("checking if an actuator of key '" + key
                + "' can be created with interface class " + actuatorClass);

        if (!actuatorKeysToFactories.containsKey(key)) {
            logger.debug("No key in map for this actuator");
            return false;
        }
        if (!actuatorKeysToFactories.get(key).canCreateActuator(key,
                actuatorClass)) {
            logger.debug("factory " + actuatorKeysToFactories.get(key)
                    + " cannot create actuator.");
            return false;
        }

        logger.debug("Can create actuator");

        return true;

    }

    /**
     * Factory method to create a new sensor for the given data type provided by
     * the resulting sensor. A sensor with this signature can only be created if
     * a call to {@link #canCreateSensor(String, Class)} returned
     * <code>true</code>.
     * 
     * @param <T>
     *            data type of the sensor to create
     * @param key
     *            key under which the sensor is registered in the configuration
     * @param dataType
     *            class object representing the data type of the sensor to
     *            create
     * @return new instance of the sensor
     * @throws IllegalArgumentException
     *             there is no sensor for the given data type with the provided
     *             key
     * @throws CoreObjectCreationException
     *             error while creating the requested sensor
     */
    public <T> Sensor<T> createSensor(String key, Class<T> dataType)
            throws IllegalArgumentException, CoreObjectCreationException {

        logger.info("Creating a sensor instance with key '" + key
                + "' and dataType '" + dataType + "'");

        if (!sensorKeysToFactories.containsKey(key)) {
            throw new IllegalArgumentException("No sensor with key '" + key
                    + "'");
        }

        CoreObjectFactory factory = sensorKeysToFactories.get(key);
        assert factory != null;
        if (!factory.canCreateSensor(key, dataType)) {
            throw new IllegalArgumentException("Factory " + factory
                    + " cannot create a sensor with key '" + key
                    + "' and dataType '" + dataType + "'.");
        }

        return factory.createSensor(key, dataType);

    }

    /**
     * Factory method to create a new actuator of the provided type.
     * 
     * @param <T>
     *            type of the actuator to create (minimally implemented
     *            interface)
     * @param key
     *            key under which the actuator is registered in the
     *            configuration
     * @param actuatorClass
     *            class object representing the type of the actuator to create
     * @return new instance of the actuator
     * @throws IllegalArgumentException
     *             there is no actuator for the given type with the provided key
     * @throws CoreObjectCreationException
     *             error while creating the requested actuator
     */
    public <T extends Actuator> T createActuator(String key,
            Class<T> actuatorClass) throws IllegalArgumentException,
            CoreObjectCreationException {

        logger.info("Creating an actuator instance with key '" + key
                + "' and type '" + actuatorClass + "'");

        if (!actuatorKeysToFactories.containsKey(key)) {
            throw new IllegalArgumentException("No actuator with key '" + key
                    + "'");
        }

        CoreObjectFactory factory = actuatorKeysToFactories.get(key);
        assert factory != null;
        if (!factory.canCreateActuator(key, actuatorClass)) {
            throw new IllegalArgumentException("Factory " + factory
                    + " cannot create an actuator with key '" + key
                    + "' and type '" + actuatorClass + "'.");
        }

        return factory.createActuator(key, actuatorClass);

    }

    /**
     * Finds the factory with the given class name.
     * 
     * @return factory with the given class name or <code>null</code> if not
     *         found
     */
    private Class<? extends CoreObjectFactory> findFactory(String className) {
        for (Class<? extends CoreObjectFactory> clazz : configuredFactoriesByClass
                .keySet()) {
            if (clazz.getName().equals(className)) {
                return clazz;
            }
        }
        return null;
    }

    /**
     * Initializes all configured sensors in their appropriate factories found
     * in {@link #configuredFactoriesByClass}. Afterwards the mapping
     * {@link #sensorKeysToFactories} is populated.
     * 
     * @param configuredSensors
     *            sensors to initialize
     * @throws ConfigurationExceptionerror
     *             initialization error
     */
    private void initializeSensors(Set<ConfiguredSensor> configuredSensors)
            throws ConfigurationException {

        // first, map every configured sensor to the factory that creates it
        Map<CoreObjectFactory, Set<SensorToConfigure>> sensorToConfigureByFactory =
                mapConfiguredCoreObjectToFactories(
                        configuredSensors,
                        new CoreObjectMapper<ConfiguredSensor, SensorToConfigure>() {

                            @Override
                            public SensorToConfigure map(ConfiguredSensor t)
                                    throws ConfigurationException {
                                return mapSensor(t);
                            }
                        });

        // then, initialize the sensors of every factory
        for (Entry<CoreObjectFactory, Set<SensorToConfigure>> factoryEntry : sensorToConfigureByFactory
                .entrySet()) {
            try {
                logger.debug("Factory " + factoryEntry.getKey() + ": "
                        + factoryEntry.getValue());
                factoryEntry.getKey().configureSensors(factoryEntry.getValue());
            } catch (IllegalArgumentException e) {
                String message =
                        "Error configuring a sensor for factory "
                                + factoryEntry.getKey();
                logger.error(message);
                throw new ConfigurationException(message, e);
            } catch (CoreObjectCreationException e) {
                String message =
                        "Error configuring a sensor for factory "
                                + factoryEntry.getKey()
                                + ". Connection of the sensor failed in the factory.";
                logger.error(message);
                throw new ConfigurationException(message, e);
            }
        }

        // if the initialization was successful, populate the map of sensor
        // keys to factory objects
        logger.info("Building map of initialized sensors");
        for (Entry<CoreObjectFactory, Set<SensorToConfigure>> factoryEntry : sensorToConfigureByFactory
                .entrySet()) {

            CoreObjectFactory factory = factoryEntry.getKey();
            logger.debug("   factory = " + factory);

            for (SensorToConfigure sensor : factoryEntry.getValue()) {

                logger.debug("      sensor = " + sensor);

                if (sensorKeysToFactories.containsKey(sensor.getKey())) {
                    throw new ConfigurationException(
                            "There is already a sensor configured with key '"
                                    + sensor.getKey() + "'. "
                                    + "Cannot initialize another sensor"
                                    + " with this key: " + sensor);
                }

                sensorKeysToFactories.put(sensor.getKey(), factory);

            }

        }

    }

    /**
     * Initializes all configured actuators in their appropriate factories found
     * in {@link #configuredFactoriesByClass}. Afterwards the mapping
     * {@link #actuatorKeysToFactories} is populated.
     * 
     * @param configuredActuators
     *            actuators to initialize
     * @throws ConfigurationExceptionerror
     *             initialization error
     */
    private void initializeActuators(Set<ConfiguredActuator> configuredActuators) {

        // map requested actuators to the factories that shall create them
        Map<CoreObjectFactory, Set<ActuatorToConfigure>> actuatorsToConfigureByFactory =
                mapConfiguredCoreObjectToFactories(
                        configuredActuators,
                        new CoreObjectMapper<ConfiguredActuator, ActuatorToConfigure>() {

                            @Override
                            public ActuatorToConfigure map(ConfiguredActuator t)
                                    throws ConfigurationException {
                                return mapActuator(t);
                            }
                        });

        // initialize actuators in each factory
        for (Entry<CoreObjectFactory, Set<ActuatorToConfigure>> factoryEntry : actuatorsToConfigureByFactory
                .entrySet()) {
            try {
                logger.debug("Factory " + factoryEntry.getKey() + ": "
                        + factoryEntry.getValue());
                factoryEntry.getKey().configureActuators(
                        factoryEntry.getValue());
            } catch (IllegalArgumentException e) {
                String message =
                        "Error configuring an actuator for factory "
                                + factoryEntry.getKey();
                logger.error(message);
                throw new ConfigurationException(message, e);
            } catch (CoreObjectCreationException e) {
                String message =
                        "Error configuring an actuator for factory "
                                + factoryEntry.getKey()
                                + ". Connection of the actuator failed "
                                + "in the factory.";
                logger.error(message);
                throw new ConfigurationException(message, e);
            }
        }

        // build a mapping of actuator keys to creating factories
        logger.info("Building map of initialized actuators");
        for (Entry<CoreObjectFactory, Set<ActuatorToConfigure>> factoryEntry : actuatorsToConfigureByFactory
                .entrySet()) {

            CoreObjectFactory factory = factoryEntry.getKey();
            logger.debug("   factory = " + factory);

            for (ActuatorToConfigure actuator : factoryEntry.getValue()) {

                logger.debug("      actuator = " + actuator);

                if (actuatorKeysToFactories.containsKey(actuator.getKey())) {
                    throw new ConfigurationException(
                            "There is already an actuator configured with key '"
                                    + actuator.getKey() + "'. "
                                    + "Cannot initialize another actuator "
                                    + "with this key: " + actuator);
                }

                actuatorKeysToFactories.put(actuator.getKey(), factory);

            }

        }

    }

    /**
     * Maps {@link ConfiguredActuator}s to {@link ActuatorToConfigure}s with
     * appropriate class checking.
     * 
     * @param confAct
     *            actuator to map
     * @return mapped actuator
     * @throws ConfigurationException
     *             error while mapping, especially class finding errors
     */
    @SuppressWarnings("unchecked")
    private ActuatorToConfigure mapActuator(ConfiguredActuator confAct)
            throws ConfigurationException {

        ActuatorToConfigure actuator = new ActuatorToConfigure();

        actuator.setKey(confAct.getKey());
        actuator.setActuatorOptions(confAct.getOptions());

        try {

            Class<?> possibleInterfaceClass =
                    Class.forName(confAct.getActuatorClassName());
            if (!Actuator.class.isAssignableFrom(possibleInterfaceClass)) {
                throw new ConfigurationException("Class "
                        + possibleInterfaceClass
                        + " does not implement interface " + Actuator.class);
            }
            actuator
                    .setInterfaceClass((Class<? extends Actuator>) possibleInterfaceClass);

            Class<?> possibleActuatorClass =
                    Class.forName(confAct.getActuatorClassName());
            if (!Actuator.class.isAssignableFrom(possibleActuatorClass)) {
                throw new ConfigurationException("Class "
                        + possibleActuatorClass
                        + " does not implement interface " + Actuator.class);
            }
            if (!actuator.getInterfaceClass().isAssignableFrom(
                    possibleActuatorClass)) {
                throw new ConfigurationException("Class "
                        + possibleActuatorClass + " does not implement its "
                        + "declared actuator interface "
                        + actuator.getInterfaceClass());
            }
            actuator
                    .setActuatorClass((Class<? extends Actuator>) possibleActuatorClass);

        } catch (ExceptionInInitializerError e) {
            throw new ConfigurationException(e);
        } catch (LinkageError e) {
            throw new ConfigurationException(e);
        } catch (ClassNotFoundException e) {
            throw new ConfigurationException(e);
        }

        return actuator;

    }

    /**
     * An interface describing a method to map one data type to another. This is
     * used to map configuration objects from {@link BonsaiManager} to instances
     * of {@link ConfiguredXcfCoreObject}.
     * 
     * @author jwienke
     * @param <T>
     *            source data type for the mapping
     * @param <U>
     *            target data type for the mapping
     */
    private interface CoreObjectMapper<T, U> {

        U map(T t) throws ConfigurationException;
    }

    /**
     * Maps all configured sensors given to the factory instances that create
     * the sensor.
     * 
     * @param <T>
     *            type of the core objects to map
     * @param configuredObjects
     *            sensors to map
     * @param mapper
     *            mapper from <code>T</code> to <code>U</code>
     * @return map that maps every given {@link ConfiguredSensor} to the factory
     *         object that can create it
     * @throws ConfigurationException
     *             on of the sensors cannot be mapped to a factory instance
     */
    private <T extends ConfiguredCoreObject, U> Map<CoreObjectFactory, Set<U>> mapConfiguredCoreObjectToFactories(
            Set<T> configuredObjects, CoreObjectMapper<T, U> mapper)
            throws ConfigurationException {

        logger.debug("Mapping ConfiguredSensors to factories. Got "
                + configuredObjects.size() + " sensors to map.");

        Map<CoreObjectFactory, Set<U>> coreObjectByFactory =
                new HashMap<CoreObjectFactory, Set<U>>();

        for (T coreObject : configuredObjects) {

            Class<? extends CoreObjectFactory> factoryClass =
                    findFactory(coreObject.getFactoryClassName());
            if (factoryClass == null) {
                throw new ConfigurationException(
                        "Cannot setup sensor with key '" + coreObject.getKey()
                                + "' because there is no factory set up "
                                + "with class '"
                                + coreObject.getFactoryClassName() + "'");
            }

            CoreObjectFactory factory =
                    configuredFactoriesByClass.get(factoryClass);
            assert factory != null : "We could find a factory with class "
                    + factoryClass + " so it must be contained in"
                    + "the map of configured factories.";
            if (!coreObjectByFactory.containsKey(factory)) {
                coreObjectByFactory.put(factory, new HashSet<U>());
            }
            logger.debug(coreObject + " belongs to " + factory);
            boolean notReplaced =
                    coreObjectByFactory.get(factory)
                            .add(mapper.map(coreObject));
            assert notReplaced;

        }

        if (logger.isDebugEnabled()) {
            for (Entry<CoreObjectFactory, Set<U>> entry : coreObjectByFactory
                    .entrySet()) {
                StringBuilder builder = new StringBuilder();
                builder.append("Mapping of sensors to factories:\n");
                builder.append("   " + entry.getKey() + ":\n");
                for (U s : entry.getValue()) {
                    builder.append("      " + s + "\n");
                }
                logger.debug(builder.toString());
            }
        }

        return coreObjectByFactory;

    }

    /**
     * Maps {@link ConfiguredSensor}s to {@link ActuatorToConfigure}s with
     * appropriate class checking.
     * 
     * @param confSensor
     *            sensor to map
     * @return mapped sensor
     * @throws ConfigurationException
     *             error while mapping, especially class finding errors
     */
    @SuppressWarnings("unchecked")
    private SensorToConfigure mapSensor(ConfiguredSensor confSensor)
            throws ConfigurationException {

        SensorToConfigure sensor = new SensorToConfigure();

        sensor.setKey(confSensor.getKey());
        sensor.setSensorOptions(confSensor.getOptions());

        try {

            sensor.setDataTypeClass(Class.forName(confSensor
                    .getDataTypeClassName()));

            Class<?> possibleSensorClass =
                    Class.forName(confSensor.getSensorClassName());
            if (!Sensor.class.isAssignableFrom(possibleSensorClass)) {
                throw new ConfigurationException("Class " + possibleSensorClass
                        + " does not implement interface " + Sensor.class);
            }
            sensor
                    .setSensorClass((Class<? extends Sensor<?>>) possibleSensorClass);

        } catch (ExceptionInInitializerError e) {
            throw new ConfigurationException(e);
        } catch (LinkageError e) {
            throw new ConfigurationException(e);
        } catch (ClassNotFoundException e) {
            throw new ConfigurationException(e);
        }

        return sensor;

    }

    /**
     * Creates the collection of user configured factories.
     * 
     * @param requestedFactories
     *            set of factory settings found in the configuration file.
     * @throws ConfigurationException
     *             error configuring a factory
     */
    private void createFactories(
            Set<ConfiguredCoreObjectFactory> requestedFactories)
            throws ConfigurationException {

        logger.debug("Creating factories: " + requestedFactories);

        if (requestedFactories.size() < 1) {
            throw new ConfigurationException(
                    "No factories requested by the user. "
                            + "This does not make sense, because then "
                            + "no sensor or actuator can be created.");
        }

        for (ConfiguredCoreObjectFactory requestedFactory : requestedFactories) {
            createFactory(requestedFactory);
        }

    }

    /**
     * Instantiates a configured factory and adds it to
     * {@link #configuredFactoriesByClass}.
     * 
     * @param requestedFactory
     *            factory to configure
     * @throws ConfigurationException
     *             error configuring the factory
     */
    private void createFactory(ConfiguredCoreObjectFactory requestedFactory)
            throws ConfigurationException {

        try {

            // first check that a class with the desired name exists and can be
            // instantiated
            Class<?> rawFactoryClass =
                    Class.forName(requestedFactory.getFactoryClassName());
            Object rawInstance = rawFactoryClass.newInstance();
            if (!(rawInstance instanceof CoreObjectFactory)) {
                throw new ConfigurationException("Class '"
                        + requestedFactory.getFactoryClassName()
                        + "' is not an instance of "
                        + CoreObjectFactory.class.getSimpleName());
            }
            CoreObjectFactory newFactory = (CoreObjectFactory) rawInstance;

            // ensure that there isn't already a configured factory of this
            // class
            Class<? extends CoreObjectFactory> factoryClass =
                    newFactory.getClass();
            if (configuredFactoriesByClass.containsKey(factoryClass)) {
                throw new ConfigurationException("A factory of class '"
                        + requestedFactory.getFactoryClassName()
                        + "' was already configured before.");
            }

            // configure the new factory
            try {
                newFactory.initialize(requestedFactory.getFactroyOptions());
            } catch (IllegalArgumentException e) {
                throw new ConfigurationException(
                        "Error configuring factory of class " + factoryClass
                                + " with the user supplied options: "
                                + requestedFactory.getFactroyOptions(), e);
            } catch (InitializationException e) {
                throw new ConfigurationException(
                        "Error initializing factory of class " + factoryClass
                                + ".", e);
            }

            // add it to the list of configured factories
            configuredFactoriesByClass.put(factoryClass, newFactory);

        } catch (ClassNotFoundException e) {
            logger.debug(e.getClass().getSimpleName(), e);
            throw new ConfigurationException(
                    "Could not find a class with name '"
                            + requestedFactory.getFactoryClassName() + "'", e);
        } catch (ExceptionInInitializerError e) {
            logger.debug(e.getClass().getSimpleName(), e);
            throw new ConfigurationException(e);
        } catch (LinkageError e) {
            logger.debug(e.getClass().getSimpleName(), e);
            throw new ConfigurationException(e);
        } catch (InstantiationException e) {
            logger.debug(e.getClass().getSimpleName(), e);
            throw new ConfigurationException(
                    "Error creating a new instance of class '"
                            + requestedFactory.getFactoryClassName()
                            + "'. Is it a top level member?", e);
        } catch (IllegalAccessException e) {
            logger.debug(e.getClass().getSimpleName(), e);
            throw new ConfigurationException(
                    "Error creating a new instance of class '"
                            + requestedFactory.getFactoryClassName()
                            + "'. Does it have a default constructor?", e);
        } catch (SecurityException e) {
            logger.debug(e.getClass().getSimpleName(), e);
            throw new ConfigurationException(
                    "Error creating a new instance of class '"
                            + requestedFactory.getFactoryClassName() + "'", e);
        }

    }

}
