/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.states;

import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.sensors.FaceIdentificationSensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.skills.abstractstates.AbstractState;
import de.unibi.airobots.bonsai.skills.abstractstates.AbstractState.ExitStatus;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.EndException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.ExecuteException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.InitException;
import de.unibi.airobots.bonsai.skills.helper.PersonManager;
import de.unibi.airobots.btl.Type;
import de.unibi.airobots.btl.data.person.PersonData;
import de.unibi.airobots.btl.data.person.PersonDataList;
import java.io.IOException;
import java.util.Date;

import org.apache.log4j.Logger;

/**
 * This state will add a new person to a PersonDataList and learn its face.
 * 
 * If a PersonDataList has been written to SceneMemory before it will be used.
 * 
 * <pre>
 * Special exit statuses: <code>error</code> if no person is standing in front
 * of the robot.
 * </pre>
 * 
 * @see PersonDataList
 * 
 * @author lkettenb
 */
public class StoreNewPerson extends AbstractState {
	/*
	 * What the robot will say.
	 */
	private static final String STATE_INTRO1 = "Please look into my eye while I tell you a bit about myself.";
	private static final String STATE_INTRO2 = "My name is ToBI and I'm from Bielefeld.";
	private static final String STATE_INTRO3 = "What I'm doing right now is trying to learn your face.";
	private static final String STATE_INTRO4 = "So please give me a few more seconds.";
	private static final String STATE_FINISHED = "Ok. From now on i am able to recognize you!";

	/*
	 * Actuators used by this state.
	 */
	private SpeechActuator speechActuator;

	/*
	 * Sensors used by this state.
	 */
	private FaceIdentificationSensor faceSensor;
	private Sensor<PersonList> personSensor;
	private Sensor<PersonDataList> personDataListSensor;
	private Sensor<PersonData> personToLearnMemorySensor;
	/**
	 * The logger.
	 */
	private Logger logger = getLogger();

	/**
	 * Additional time to learn person.
	 */
	private static final int TIME_TO_LEARN = 10000;

	/**
	 * List with person data. Will be read from and saved to the scene memory.
	 */
	private PersonDataList personDataList;

	/**
	 * Person which has to be stored.
	 */
	private PersonData personToLearn;

	/**
	 * True, if person data was read from memory.
	 */
	private boolean personDataReadFromMem = false;
	private Date start;

	/**
	 * Error message if memory cannot be read.
	 */
	private static final String ERROR_MEM = "Unable to access memory. "
			+ "PersonDataList will be overwritten. ";

	/**
	 * Constructor.
	 * 
	 * @param aStateMachine
	 *            The state machine to use.
	 */
	public StoreNewPerson(SkillStateMachine aStateMachine) {
		super(aStateMachine);
	}

	@Override
	public ExitStatus init() throws InitException {
		// Initialize actuators
		speechActuator = getManagedActuator("SpeechActuator");
		// Initialize sensors
		faceSensor = (FaceIdentificationSensor) getManagedSensor("FaceIdentificationSensor");
		personSensor = getManagedSensor("PersonSensor");
		personDataListSensor = getManagedSensor("knownPersonsMemorySensor");

		if (personDataListSensor.hasNext()) {
			try {
				personDataList = personDataListSensor.readLast();
			} catch (IOException ex) {
				logger.warn(ERROR_MEM + ex.getMessage());
			} catch (InterruptedException ex) {
				logger.warn(ERROR_MEM + ex.getMessage());
			}
			personDataReadFromMem = true;
		} else {
			personDataList = new PersonDataList();
		}
		personToLearn = PersonManager
				.getNextPersonInFront(personSensor);

		if (personToLearn == null || personDataList == null) {
			return ExitStatus.ERROR;
		}

		start = new Date();

		personToLearn.setId(personDataList.size());
		// start learning
		try {
			faceSensor.activate();
			faceSensor.startLearning(personToLearn.getId());
		} catch (IOException e) {
			e.printStackTrace();
			return ExitStatus.FAIL;
		}

		return ExitStatus.SUCCESS;
	}

	@Override
	public ExitStatus execute() throws ExecuteException {
		// do until learntime is elapsed
		if ((new Date().getTime() - start.getTime()) > TIME_TO_LEARN) {
			return ExitStatus.SUCCESS;
		} else {
			setProcessingStatus(getWaitProcessingMessage());
			return ExitStatus.SUCCESS;
		}

	}

	@Override
	public ExitStatus end() throws EndException {
		if (isForcedToEnd()) {
			say("Please give me a few more seconds");
			try {
				Thread.sleep(TIME_TO_LEARN
						- (new Date().getTime() - start.getTime()));
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				return ExitStatus.FAIL;
			}
		}
		try {
			faceSensor.stopLearning();
			faceSensor.deactivate();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return ExitStatus.FAIL;
		}
		return ExitStatus.SUCCESS;
	}

	/**
	 * Use speech actuator to say something and catch IO exception.
	 * 
	 * @param text
	 *            Text to be said.
	 */
	private void say(String text) {
		try {
			speechActuator.say(text);
		} catch (IOException ex) {
			// Not so bad. The robot just says nothing.
			logger.warn(ex.getMessage());
		}
	}

}
