/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.xcf.configuration;

import java.util.Map;

/**
 * Wraps one configured xcf core object with its settings.
 * <p>
 * TODO:
 * <ul>
 * <li>Provide reasonable defaults for all options in all subclasses.</li>
 * </ul>
 * <p>
 * Every subclass must be default constructible.
 * 
 * @author jwienke
 */
public abstract class ConfiguredXcfCoreObject {

    /**
     * Key under which the core object is registered in the configuration.
     */
    private String key = "";

    /**
     * Parse options for this configuration.
     * 
     * @param options
     *            options in key value form
     * @throws IllegalArgumentException
     *             options not suitable for creating such a configuration object
     */
    public abstract void parseOptions(Map<String, String> options)
            throws IllegalArgumentException;

    /**
     * Returns the names of remote servers that need to be initialized for this
     * configured object to work.
     * 
     * Implement this in subclasses if a remote server is needed for the object.
     * 
     * @return array of remote server names
     */
    public String[] getRemoteServersToInitialize() {
        return new String[0];
    }

    /**
     * Returns the names of subscribers that need to be initialized for this
     * configured object to work.
     * 
     * Implement this in subclasses if a subscriber is needed for the object.
     * 
     * @return array of remote server names
     */
    public String[] getSubscribersToInitialize() {
        return new String[0];
    }

    /**
     * Returns the names of active memories that need to be initialized for this
     * configured object to work.
     * 
     * Implement this in subclasses if an active memory is needed for the
     * object.
     * 
     * @return array of remote server names
     */
    public String[] getActiveMemoriesToInitialize() {
        return new String[0];
    }

    /**
     * Returns the key under which the core object was registered in the
     * configuration.
     * 
     * @return key string
     */
    public String getKey() {
        return key;
    }

    /**
     * Setter for the key under which the core object was configured by the
     * user.
     * 
     * @param key
     *            new key
     */
    public void setKey(String key) {
        this.key = key;
    }

}
